#pragma module TRACEMEM "V1.1-0"
/*******************************************************************************
*									       *
*	TRACEMEM	Trace use of malloc() and free().		       *
*									       *
* This routine provides a means for identifying memory leaks in programs which *
* use malloc() to allocate blocks of memory.				       *
*									       *
* To use these routines, the calling program's source code must include the    *
* following statement:							       *
*									       *
*	#include "TRACEMEM.H"						       *		
*									       *
* This causes the program to call trace_malloc() and trace_free() instead of   *
* malloc() and free().							       *
*									       *
* The trace_malloc() routine maintains two data structures: a lookup tree of   *
* source code locations where malloc() is called, and a hash table of all      *
* allocated pointers (tied back to the source code references).		       *
*									       *
* The trace_free() routine uses the hash table to locate each freed pointer;   *
* when a pointer is passed to trace_free() it is removed from the hash table   *
* and the statistics for the corresponding source code reference are adjusted  *
* accordingly.								       *
*									       *
* Periodically the lookup tree of source code references is dumped out in text *
* form so that memory leaks can be investigated.  The frequency of dumps is    *
* determined by the logical name TRACEMEM_DUMP_INTERVAL which can be either an *
* integer value or a delta-time.  If an integer 'N' is specified, the table is *
* dumped every 'N' calls to trace_malloc().  If a delta-time 'HH:MM:SS' is     *
* specified, a timer AST will set a flag so that the next call to trace_malloc *
* results in the table being dumped.					       *
*									       *
* In addition, an exit handler is established to dump out the table when the   *
* program exits.  (This might not be 100% reliable, e.g. if the program exits  *
* because of an INSVIRMEM error or if the process is deleted.)		       *
*									       *
* The dump will be written to the file identified by the logical name          *
* TRACEMEM_DUMP_FILE if it exists; if the logical does not exist the dump will *
* be written to SYS$OUTPUT.						       *
*									       *
* If neither TRACEMEM_DUMP_INTERVAL nor TRACEMEM_DUMP_FILE are defined before  *
* the first time a TRACEMEM routine is called. no statistics will be gathered. *	
*									       *
*	Created 17-Jan-2011 by Jeremy Begg, VSM Software Services Pty Ltd.     *
*									       *
* Modifications								       *
* -------------								       *
* 16-Jun-2011	J.Begg		Tweak it for VAX support.		       *
*				Clean up some of the output.		       *
*				Add logic to inhibit statistics if no logicals *
*				are defined at startup.			       *
*									       *
*******************************************************************************/

#include <assert.h>
#include <builtins.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <descrip.h>
#include <jpidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <rms.h>
#include <ssdef.h>
#include <starlet.h>
#include <lib$routines.h>


/*******************************************************************************
*	DATA STRUCTURE DEFINITIONS
*******************************************************************************/

#ifdef __VAX
#define __MEMSET(dest, value, count)  _MOVC5(0, (const char *)(dest), (value), (count), (char *)(dest))
typedef long int Big_Counter;		/* No "big" support on VAX */
typedef struct { unsigned long int lo; long int hi; } VMS_Bintim;
#else
typedef __int64 Big_Counter;
typedef __int64 VMS_Bintim;
#endif

/*
*  Source code reference lookup tree
*/
typedef struct _source_lookup {
	struct _source_lookup	*left, *right;	/* Binary search tree         */
	struct _pointer_info	*head, *tail;	/* Doubly-linked list of active pointers */
	char 			*source_line;	/* Module name & line number  */
	int			refcount;	/* No. of times this source line has been spotted */
	int			active_count;	/* No of active pointers allocated by this line */
	Big_Counter		total_alloc;	/* Total space ever allocated by this source line */
	Big_Counter		current_alloc;	/* Best-guess current space allocated to this source line */
	} *p_source_lookup_t;

/*
*  Details for each allocated pointer
*/
typedef struct _pointer_info {
	struct _pointer_info	*prev, *next;	/* Doubly-linked list in hash table */
	struct _pointer_info	*a_prev, *a_next; /* Doubly-linked list of active pointers */
	struct _source_lookup	*source_ref;	/* Where it was allocated */
	void			*pointer;	/* The pointer */
	unsigned long int	alloc_size;	/* Size of block it points to */
	} *p_pointer_info_t;
	
/*
*  All pointers are entered into a hash table.  The hash is indexed by bits
*  19-4 of the pointer value.  (This is a 16-bit index value, or 64K entries.)
*  Each hash table entry points to a doubly-linked list of pointers, ordered
*  by pointer value, and some statistics.
*/
typedef struct _hash_entry {
	struct _pointer_info	*head, *tail;	/* Doubly-linked hash list */
	long int		count;		/* Count of pointers in this hash */
	} hash_entry_t;


/*******************************************************************************
*	MODULE-LOCAL STATIC DATA
*******************************************************************************/

#define HASH_TABLE_SIZE		65536

static hash_entry_t		hash_table[HASH_TABLE_SIZE];	/* The hash table */
static p_source_lookup_t	source_lines=NULL;		/* Source lookup tree */

static volatile int		malloc_calls=0;			/* Number of calls to trace_malloc() */
static volatile int		free_calls=0;			/* Number of calls to trace_free() */
static volatile int		malloc_intervals=0;		/* Interval counter */

static VMS_Bintim		dump_interval;			/* Dump interval (VMS binary delta-time) */
static volatile int		dump_flag=0;			/* Dump at next trace_malloc() call */
static char			dump_file[256];			/* Filespec to open */
static int			dump_efn=0;			/* EFN for timer; if 0, use iteration count instead */

static int			no_tracing=0;			/* Assume we will be tracing */

#ifdef __VAX
#define dump_at dump_interval.lo
#else
#define dump_at dump_interval
#endif

static struct FAB		dumpfile_FAB;
static struct RAB		dumpfile_RAB;

static void trace_dump();			/* forward declaration */

/*******************************************************************************
*	DUMP FILE OUTPUT						       *
*									       *
* The routines in this section are responsible for generating output to the    *
* dump file.								       *
*									       *
*******************************************************************************/

static void trace_dumpAST()		/* Set a flag to request a trace dump	*/
{
/*
*  Set a flag to ensure the next call to trace_malloc() will generate some
*  output to the dump file.  Then reset the timer for next time.
*/
    dump_flag = 1;
    sys$setimr(dump_efn, &dump_interval, trace_dumpAST);
    return;
}


static void trace_dump_write (char *text)
{
    if (text) {
	dumpfile_RAB.rab$l_rbf = text;
	dumpfile_RAB.rab$w_rsz = strlen(text);
    }
    else {
	dumpfile_RAB.rab$l_rbf = (char *) &text;
	dumpfile_RAB.rab$w_rsz = 0;
    }
    sys$put (&dumpfile_RAB);
}

/*******************************************************************************
*	INITIALISATION							       *
*									       *
* Initialise the hash array						       *
* Set the dump interval							       *
* Open the output file							       *
*									       *
*******************************************************************************/


void trace_initialise()
{
    int i, ret_status;
    char buffer[256];
    struct dsc$descriptor_s bufdescr = {0, DSC$K_DTYPE_T, DSC$K_CLASS_S, buffer};
    short int buflen;
    char *filename_logical=0, *interval_logical=0;
    float seconds;

/*
*  Check the TRACEMEM logicals.  If none are defined, no tracing is to be performed.
*/
    filename_logical = getenv("TRACEMEM_DUMP_FILE");
    interval_logical = getenv("TRACEMEM_DUMP_INTERVAL");
    if ((filename_logical == NULL) && (interval_logical == NULL)) {
	no_tracing = 1;
	return;
    }

/*
*  Initialise the hash table
*/
    for (i=0; i < HASH_TABLE_SIZE; i++) {
	hash_table[i].head = NULL;
	hash_table[i].tail = NULL;
	hash_table[i].count = 0;
    }


/*
*  Set the output file name
*/
    if (filename_logical) {
	strcpy(dump_file, filename_logical);
	no_tracing = 0;
    }
    else {
	strcpy(dump_file, "SYS$OUTPUT:");
    }

/*
*  Open the output file
*/
    dumpfile_FAB = cc$rms_fab;
    dumpfile_FAB.fab$b_fac = FAB$M_PUT;
    dumpfile_FAB.fab$l_fna = dump_file;
    dumpfile_FAB.fab$b_fns = strlen(dump_file);
    dumpfile_FAB.fab$l_fop = FAB$M_SQO;
    dumpfile_FAB.fab$b_org = FAB$C_SEQ;
    dumpfile_FAB.fab$b_rat = FAB$M_CR;
    dumpfile_FAB.fab$b_rfm = FAB$C_VAR;
    dumpfile_FAB.fab$b_shr = FAB$M_SHRGET;

    dumpfile_RAB = cc$rms_rab;
    dumpfile_RAB.rab$l_fab = &dumpfile_FAB;

    ret_status = sys$create(&dumpfile_FAB);
    if (ret_status & 1) ret_status = sys$connect(&dumpfile_RAB);

/*
*  Get the name of the main program so we can put something in the dump file
*/
    strcpy(buffer, "TRACEMEM: Tracking memory allocation for ");
    bufdescr.dsc$w_length = 255 - strlen(buffer);
    bufdescr.dsc$a_pointer += strlen(buffer);
    ret_status = lib$getjpi(&JPI$_IMAGNAME, 0, 0, 0, &bufdescr, &buflen);
    if (ret_status & 1) {
	bufdescr.dsc$a_pointer[buflen] = '\0';
	trace_dump_write(buffer);
    }


/*
*  Determine the dump interval
*/
    __MEMSET(&dump_interval, 0, sizeof(dump_interval));
    if (interval_logical) {
	sprintf(buffer, "TRACEMEM: TRACEMEM_DUMP_INTERVAL=%s", interval_logical);  trace_dump_write(buffer);
	if (strchr(interval_logical,':') == NULL) {
	    /* TRACEMEM_DUMP_INTERVAL is an integer */
#ifdef __VAX
	    dump_at = atol(interval_logical);
	    if (dump_at > 0) {
		sprintf(buffer, "TRACEMEM: Trace information will be dumped every %d iterations", dump_at);
#else
	    dump_at = atoq(interval_logical);
	    if (dump_at > 0) {
		sprintf(buffer, "TRACEMEM: Trace information will be dumped every %Ld iterations", dump_at);
#endif
		trace_dump_write(buffer);
	    }
	}
	else {
	    /* TRACEMEM_DUMP_INTERVAL is a VMS delta-time */
	    struct dsc$descriptor_s timbuf = {strlen(interval_logical), DSC$K_DTYPE_T, DSC$K_CLASS_S, interval_logical};
	    ret_status = sys$bintim(&timbuf, &dump_interval);
	    if (ret_status & 1) {
		/* Set up a timer to trigger at the specified interval */
		lib$get_ef(&dump_efn);
		lib$cvtf_from_internal_time(&LIB$K_DELTA_SECONDS_F, &seconds, &dump_interval);
		sprintf(buffer, "TRACEMEM: Trace information will be dumped every %1.0f seconds (approx)", seconds);
		trace_dump_write(buffer);
		ret_status = sys$setimr(dump_efn, &dump_interval, trace_dumpAST);
		if (!(ret_status & 1)) {
		    sprintf(buffer, "TRACEMEM: ERROR %08X calling SYS$SETIMR, using default dump interval instead", ret_status);
		    trace_dump_write(buffer);
		    lib$free_ef (&dump_efn);
		    dump_efn = 0;
		    __MEMSET(&dump_interval, 0, sizeof(dump_interval));
		}
	    }
	    else {
		trace_dump_write("TRACEMEM: TRACE_DUMP_INTERVAL specifies invalid VMS delta time");
		__MEMSET (&dump_interval,0,sizeof(dump_interval));
	    }
	}
    }

#ifdef __VAX
    if ((dump_interval.lo == 0) && (dump_interval.hi == 0)) {
	dump_at = 500;
	sprintf(buffer, "TRACEMEM: Using default dump interval of %d iterations", dump_at);
#else
    if (dump_interval == 0) {
	dump_at = 500;
	sprintf(buffer, "TRACEMEM: Using default dump interval of %Ld iterations", dump_at);
#endif
	trace_dump_write(buffer);
    }

    return;
}

/*******************************************************************************
*	SOURCE LINE LOOKUP						       *
*									       *
* This routine does a binary search of the source lookup tree and returns a    *
* pointer to an existing reference or a pointer to a new reference (i.e. when  *
* the source line in question has not been seen before).		       *
*									       *
*******************************************************************************/

static p_source_lookup_t lookup_source_line (p_source_lookup_t *source_ref
					    ,char *line_num)
{
    p_source_lookup_t s;
    int c;

/*
*  If *source_ref is NULL, the line has not been seen before
*/
    if (*source_ref == NULL) {
	s = malloc(sizeof(*s));
	__MEMSET (s, 0, sizeof(*s));
	*source_ref = s;
	return s;
    }

/*
*  Otherwise, do a string comparison on the line and take the appropriate branch
*/
    s = *source_ref;
    c = strcmp(s->source_line, line_num);
    if (c == 0) return s;
    if (c < 0) return lookup_source_line (&s->right, line_num);
    return lookup_source_line (&s->left, line_num);
}

/*******************************************************************************
*	HASH TABLE MAINTENANCE						       *
*									       *
* Two routines are provided for manipulating the hash table:		       *
*									       *
*	p_pointer_info_t hash_store (void  *ptr, p_source_lookup_t source, __size_t size) *
*	p_pointer_info_t hash_remove (void *ptr)			       *
*									       *
*******************************************************************************/

static p_pointer_info_t hash_store (void *ptr, p_source_lookup_t source, __size_t size)
{
    p_pointer_info_t p;
    unsigned short int ptr_idx;

/*
*  Create a new pointer_info structure for this pointer
*/
    p = malloc(sizeof(*p));
    p->prev = NULL;
    p->next = NULL;
    p->a_prev = NULL;
    p->a_next = NULL;
    p->source_ref = source;
    p->pointer = ptr;
    p->alloc_size = size;

    
/*
*  Get the hash table entry for this pointer value
*/
    ptr_idx = (((long int) ptr) & 0x000FFFF0) >> 4;

/*
*  Insert the new pointer at the appropriate place in the list
*/
    if (hash_table[ptr_idx].count == 0) {
	/* First pointer in this list */
	hash_table[ptr_idx].head = p;
	hash_table[ptr_idx].tail = p;
	hash_table[ptr_idx].count = 1;
    }
    else {
	/* Scan the list looking for the insertion point */
	if (ptr < hash_table[ptr_idx].head->pointer) {
	    p->next = hash_table[ptr_idx].head;
	    p->next->prev = p;
	    hash_table[ptr_idx].head = p;
	}
	else if (ptr > hash_table[ptr_idx].tail->pointer) {
	    p->prev = hash_table[ptr_idx].tail;
	    p->prev->next = p;
	    hash_table[ptr_idx].tail = p;
	}
	else {
	    p_pointer_info_t hp = hash_table[ptr_idx].head;
	    while (ptr > hp->pointer) hp = hp->next;
	    p->next = hp;
	    p->prev = hp->prev;
	    hp->prev->next = p;
	    hp->prev = p;
	}
	hash_table[ptr_idx].count++;
    }

    return p;
}


static p_pointer_info_t hash_remove (void *ptr)
{
    p_pointer_info_t p;
    p_source_lookup_t s;
    unsigned short int ptr_idx;

/*
*  Look for this pointer in the relevant hash list
*/
    ptr_idx = (((long int) ptr) & 0x000FFFF0) >> 4;
    p = hash_table[ptr_idx].head;
    if (p == NULL) return NULL;		/* We haven't seen this pointer before */

    while (p != NULL) {
	if (p->pointer == ptr) break;	/* We've found the pointer */
	p = p->next;
    }

/*
*  If we've found the pointer, remove it from the hash list
*/
    if (p != NULL) {
	if (hash_table[ptr_idx].head == p) {
	    hash_table[ptr_idx].head = p->next;
	    if (hash_table[ptr_idx].head) hash_table[ptr_idx].head->prev = NULL;
	}
	else {
	    p->prev->next = p->next;
	}
	if (hash_table[ptr_idx].tail == p) {
	    hash_table[ptr_idx].tail = p->prev;
	    if (hash_table[ptr_idx].tail) hash_table[ptr_idx].tail->next = NULL;
	}
	else {
	    p->next->prev = p->prev;
	}
	hash_table[ptr_idx].count--;
    }

    return p;
}

/*******************************************************************************
*	GLOBAL ROUTINES							       *
*									       *
*	void *trace_malloc(__size_t size, char *filespec, char *line)	       *
*	void trace_free(void *__ptr)					       *
*									       *
*******************************************************************************/

void *trace_malloc (__size_t size,		/* Numer of bytes to allocate */
		    char *filespec,		/* Source file where called   */
		    int  line)			/* Source line where called   */
{

    int  enable_AST = 0;

    char *m=NULL;	/* Module name */
    char *c;
    char lineref[256];

    p_source_lookup_t	source;
    p_pointer_info_t	p;

    void *ptr;					/* The newly-allocated memory */

/*
*  If we are not tracing, just call malloc() and return
*/
    if (no_tracing) return malloc(size);


/*
*  This code is not AST-reentrant, so disable ASTs if necessary.
*/
    if (lib$ast_in_prog() == 0) {
	/* Not running at AST level, so disable ASTs */
	enable_AST = (sys$setast(0) == SS$_WASSET);
    }

/*
*  If this is the first time through, initialise the dumping
*/
    if (malloc_calls == 0) trace_initialise();

/*
*  Keep some basic stats
*/
    malloc_calls++;

/*
*  See if we should dump.  (If dump_efn != 0, a timer AST triggers the dump.)
*/
    if ((dump_efn == 0) && (malloc_intervals++ >= dump_at)) {
	dump_flag = 1;
	malloc_intervals = 1;
    }

/*
*  Remove device & directory from filename.  There could be up to two ]'s.
*  Then combine with supplied line number to build a source code reference.
*/
    if (filespec) {
	for (m=filespec; (*m != '\0') && (*m != ']'); m++);
	if (*m == ']') {
	    for (c = m++; (*c != '\0') && (*c != ']'); c++);
	    if (*c == ']') m = c+1;
	}
	sprintf(lineref, "%s/%05d", m, line);
    }
    else
	sprintf(lineref, "<unknown>/%5d", line);

/*
*  Find this source code reference in the lookup tree
*/
    source = lookup_source_line (&source_lines, lineref);
    if (source->source_line == NULL) {
	/* This is the first time we've seen this line */
	source->source_line = malloc(strlen(lineref)+1);
	strcpy(source->source_line, lineref);
    }

/*
*  Update stats for this source code reference
*/
    source->refcount++;
    source->active_count++;
    source->current_alloc += size;
    source->total_alloc += size;

/*
*  Do the allocation
*/
    ptr = malloc(size);

/*
*  Add this pointer to the hash table and to the list of active pointers
*  for this source line
*/
    p = hash_store (ptr, source, size);
    if (source->head == NULL) {
	/* This is the first pointer for this source code reference */
	source->head = p;
	source->tail = p;
	assert(source->active_count == 1);
    }
    else if (source->head->pointer > ptr) {
	/* Add the new pointer at the head of the list */
	p->a_next = source->head;
	source->head->a_prev = p;
	source->head = p;
    }
    else if (source->tail->pointer < ptr) {
	/* Add the new pointer at the tail of the list */
	p->a_prev = source->tail;
	source->tail->a_next = p;
	source->tail = p;
    }
    else {
	/* Insert the active pointer into the list somewhere */
	p_pointer_info_t ap;
	ap = source->head;
	while (ap->pointer < ptr) ap = ap->a_next;
	p->a_next = ap;
	p->a_prev = ap->a_prev;
	ap->a_prev->a_next = p;
	ap->a_prev = p;
    }

/*
*  If the dump interval says so, dump the info
*/
    if (dump_flag == 1) {
	trace_dump();
	dump_flag = 0;
    }

/*
*  Re-enable ASTs if we need to
*/
    if (enable_AST) sys$setast(1);

/*
*  Return a pointer to the block requested by the calling program
*/
    return ptr;
}


void trace_free (void *__ptr)			/* Pointer being freed	      */
{
    p_source_lookup_t s;
    p_pointer_info_t p;
    int enable_AST = 0;

/*
*  If we are not tracing, simply call free() and return
*/
    if (no_tracing) {
	free(__ptr);
	return;
    }

/*
*  This code is not AST-reentrant, so disable ASTs if necessary.
*/
    if (lib$ast_in_prog() == 0) {
	/* Not running at AST level, so disable ASTs */
	enable_AST = (sys$setast(0) == SS$_WASSET);
    }

/*
*  Keep some basic stats
*/
    free_calls++;

/*
*  Remove the pointer from the hash table
*/
    p = hash_remove (__ptr);

/*
*  If the pointer was in the table, adjust source code statistics
*  and remove the pointer from the list of active pointers.
*/
    if (p != NULL) {
	s = p->source_ref;
	s->active_count--;
	s->current_alloc -= p->alloc_size;
	if (s->head == p) {
	    s->head = p->a_next;
	    if (s->head) s->head->a_prev = NULL;
	}
	else {
	    p->a_prev->a_next = p->a_next;
	}
	if (s->tail == p) {
	    s->tail = p->a_prev;
	    if (s->tail) s->tail->a_next = NULL;
	}
	else {
	    p->a_next->a_prev = p->a_prev;
	}
	free(p);
    }

/*
*  Release the allocated memory
*/
    free(__ptr);

/*
*  Re-enable ASTs if we need to, and return
*/
    if (enable_AST) sys$setast(1);
    return;
}

/*******************************************************************************
*	DUMP ROUTINES							       *
*									       *
* The routines here dump out the pointer allocations by source line reference, *
* either at intervals specified by TRACEUMP_INTERVAL or upon request by the    *
* calling program.							       *
*									       *
*******************************************************************************/

static const $DESCRIPTOR (FAO_HDR_1, "TRACEMEM: Memory Allocations by Source Line Reference as at !%D");
#ifdef __VAX
static const $DESCRIPTOR (FAO_STATS, "!10SL  !10SL  !10SL  !10SL  !AZ");
#else
static const $DESCRIPTOR (FAO_STATS, "!10SL  !10SL  !10SQ  !10SQ  !AZ");
#endif
static const $DESCRIPTOR (FAO_TRAIL, "Total malloc() calls = !SL      Total free() calls = !SL");

static void trace_dump_VM (struct dsc$descriptor *stats, int userarg)
{
    dumpfile_RAB.rab$l_rbf = stats->dsc$a_pointer;
    dumpfile_RAB.rab$w_rsz = stats->dsc$w_length;
    sys$put (&dumpfile_RAB);
}

static void trace_dump_source (p_source_lookup_t source)
{
    char buffer[256];
    struct dsc$descriptor_s bufdescr = {0, DSC$K_DTYPE_T, DSC$K_CLASS_S, buffer};
    short int buflen;

    if (source == NULL) return;
    if (source->left) trace_dump_source(source->left);

    bufdescr.dsc$w_length = 255;
    sys$fao (&FAO_STATS, &buflen, &bufdescr, source->refcount, source->active_count, source->total_alloc, source->current_alloc, source->source_line);
    if (buflen > 255) buflen = 255;
    buffer[buflen] = '\0';
    trace_dump_write(buffer);

    if (source->right) trace_dump_source(source->right);
    return;
}

void trace_dump()
{
    /* Binary scan of the source line reference table */

    int vm_trace;
    char buffer[256];
    struct dsc$descriptor_s bufdescr = {0, DSC$K_DTYPE_T, DSC$K_CLASS_S, buffer};
    short int buflen;

/*
*  Write a heading to the file
*/
    trace_dump_write ("");
    bufdescr.dsc$w_length = 255; sys$fao (&FAO_HDR_1, &buflen, &bufdescr, 0); buffer[buflen] = '\0'; trace_dump_write(buffer);
    trace_dump_write("    Refcnt   ActiveCnt    TotSpace    CurSpace  Source");
    trace_dump_write("----------  ----------  ----------  ----------  --------------------------------");

/*
*  Scan the source line references, dumping each as we go
*/
    trace_dump_source(source_lines);

/*
*  Write out a trailer
*/
    trace_dump_write ("");
    bufdescr.dsc$w_length = 255; sys$fao (&FAO_TRAIL, &buflen, &bufdescr, malloc_calls, free_calls); buffer[buflen] = '\0'; trace_dump_write(buffer);
    trace_dump_write ("--------------------------------------------------------------------------------");
    vm_trace = 0; lib$show_vm(&vm_trace, trace_dump_VM, 0);
    vm_trace = 4; lib$show_vm(&vm_trace, trace_dump_VM, 0);
    sys$flush (&dumpfile_RAB);
    return;
}
